/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.data;

import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeBaseValue;
import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeListValue;
import de.bsvrz.dav.daf.communication.dataRepresentation.data.DataFactory;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ResultValue;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;

/**
 * Bietet eine Kapselung der Protokolldaten in ein Interface, wie es Dav-Daten benutzen. Anders als Dav-Daten sind diese Daten nicht streng an die verwendete
 * Attributgruppe bzw. das verwendete Attribut gebunden (sofern überhaupt eines existiert). Je nach Anwendungsfall bieten sich daher auch Abfragen nach {@link
 * #isDefined()}, {@link #isPlain()} usw. an, auch wenn die Attributgruppe diese Daten eigentlich nicht unterstützen würde. Besonders wichtig ist dies bei
 * Fehlermeldungen von PuA (wie z.B. einer Division durch 0, oder einem "Keine Daten"-Datensatz), da hier die Meldung unabhängig vom Attributtyp über {@link
 * #asTextValue()} abgerufen werden kann.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
public final class AttributeGroupProtocolData extends ProtocolData {

	/** Gekapselte Attributgruppe */
	private final AttributeGroup _attributeGroup;

	/** Datenmodell */
	private final DataModel _dataModel;

	/** Wert des Datums (von PuA) */
	private final ResultValue _value;

	/** Hier wird eine Text-Meldung von PuA gespeichert (z.B. "Keine Daten") */
	private final String _atomicString;

	/** Liste der Attribute in dieser ATG */
	private final List<Attribute> _attributeList;

	/** Kind-Attribute (werden zum einfachen Zugriff bereits im Konstruktor angelegt) */
	private final ArrayList<Data> _childrenByIndex;

	/** Kind-Attribute nach Bezeichnung (werden zum einfachen Zugriff bereits im Konstruktor angelegt) */
	private final Map<String, AttributeProtocolData> _attributeMap;

	/**
	 * Erstellt eine neue Instanz
	 *
	 * @param attributeGroup Attribut-Gruppe
	 * @param dataModel      Datenmodell
	 * @param value          Daten von PuA
	 */
	public AttributeGroupProtocolData(final AttributeGroup attributeGroup, final DataModel dataModel, final ResultValue value) {
		_attributeGroup = attributeGroup;
		_dataModel = dataModel;
		_value = value;
		if(value.isAtomar()) {
			_atomicString = value.getValue().toString();
			_attributeMap = null;
			_attributeList = null;
			_childrenByIndex = null;
			return;
		}
		_atomicString = null;
		final List<ResultValue> children = _value.getChildren();
		_attributeMap = new HashMap<>(children.size());
		_childrenByIndex = new ArrayList<>(children.size());
		_attributeList = _attributeGroup.getAttributes();
		for(int i = 0, attributeListSize = _attributeList.size(); i < attributeListSize; i++) {
			final Attribute attribute = _attributeList.get(i);
			final AttributeProtocolData data = new AttributeProtocolData(attribute, dataModel, value.getChildren().get(i));
			_attributeMap.put(
					attribute.getName(), data
			);
			_childrenByIndex.add(data);
		}
	}

	public Data createUnmodifiableCopy() {
		final ByteArrayOutputStream byteBuffer = new ByteArrayOutputStream();
		final DataOutputStream out = new DataOutputStream(byteBuffer);
		try {
			for(final Attribute attribute : _attributeList) {

				final AttributeBaseValue attributeBaseValue = new AttributeListValue(_dataModel, attribute);
				attributeBaseValue.writeValue(out);
			}
		}
		catch(IOException ex) {
			throw new IllegalStateException("Fehler beim Erstellen einer Kopie: ", ex);
		}
		finally {
			try {
				out.close();
			}
			catch(IOException ignored) {
			}
		}
		final byte[] dataBytes = byteBuffer.toByteArray();
		return DataFactory.forVersion(1).createUnmodifiableData(_attributeGroup, dataBytes);
	}

	public String getName() {
		return _attributeGroup.getName();
	}

	public AttributeType getAttributeType() {
		return null;
	}

	public boolean isDefined() {
		if(isList()) return true;

		// Fehlerergebnisse und keine Ergebnisse sind generell nicht definiert
		final ExpressionResult value = _value.getValue();
		if(value != null) {
			if(value.getType() == ExpressionResult.ResultType.NONE
					|| value.getType() == ExpressionResult.ResultType.ERROR
					|| value.getType() == ExpressionResult.ResultType.NO_CHANGE) {
				return false;
			}
		}

		return _value.hasData();
	}

	@Override
	public boolean isNoChange() {
		ExpressionResult value = _value.getValue();
		return value != null && value.getType() == ExpressionResult.ResultType.NO_CHANGE;
	}

	public boolean isList() {
		return _attributeMap != null;
	}

	public boolean isArray() {
		return false;
	}

	public boolean isPlain() {
		return !isList();
	}

	public Data getItem(final String itemName) {
		if(!isList()) {
			throw new UnsupportedOperationException(getName() + MESSAGE_NOT_LIST);
		}
		return _attributeMap.get(itemName);
	}

	public TextValue asTextValue() {
		if(isList()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
		return new TextValue() {
			public String getText() {
				return _atomicString;
			}

			public String getValueText() {
				return _atomicString;
			}

			public String getSuffixText() {
				return "";
			}

			public void setText(final String text) {
				throw new UnsupportedOperationException(ProtocolData.MESSAGE_NO_CHANGE);
			}
		};
	}

	public TimeValue asTimeValue() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
	}

	public NumberValue asScaledValue() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
	}

	public NumberValue asUnscaledValue() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
	}

	public ReferenceValue asReferenceValue() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
	}

	public Array asArray() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
	}

	public TextArray asTextArray() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
	}

	public TimeArray asTimeArray() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
	}

	public NumberArray asScaledArray() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
	}

	public NumberArray asUnscaledArray() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
	}

	public ReferenceArray asReferenceArray() {
		throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
	}

	@Override
	public boolean isError() {
		if(isList()) return false;

		// Fehlerergebnisse und keine Ergebnisse sind generell nicht definiert
		final ExpressionResult value = _value.getValue();
		if(value != null) {
			if(value.getType() == ExpressionResult.ResultType.ERROR) {
				return true;
			}
		}

		return false;
	}

	@Override
	protected List<Data> getChildrenList() {
		return Collections.unmodifiableList(_childrenByIndex);
	}
}
