/*
 * Copyright 2016 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.functions;

import antlr.collections.AST;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.ExpressionResultAndState;
import de.bsvrz.pua.prot.processing.util.TempAttribut;
import de.bsvrz.pua.prot.util.ErrorMessageBuilder;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ExpressionTree;
import de.bsvrz.pua.prot.util.attributes.RealElement;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;

import java.util.EnumSet;
import java.util.Hashtable;
import java.util.List;

/**
 * Implementierung einer PuA-Funktion, die einen Wahrheitswert abhängig vom Argument ermittelt.
 *
 * @author Kappich Systemberatung, Aachen
 * @version $Revision: 0000 $
 */
public abstract class UnaryPredicate implements ExpressionInterface {
	/**
	 * Bestimmt das Ergebnis der PuA-Funktion abhängig vom Argumentwert.
	 * @param argumentValue  Wert des Arguments
	 * @return Ergebnis der PuA-Funktion
	 */
	protected abstract boolean getResult(ExpressionResultAndState argumentValue);

	/**
	 * Ermittelt das Funktions-Ergebnis.
	 * @param argumentExpressions Liste mit den Eingabeparametern
	 * @param realElements   Mapping der realen Elemente zu den ihnen zugeordeten Werten.
	 * @param tempElements   Mapping der temporären Attribute zu den ihnen zugeordneten Werten.
	 * @param pseudoObjectIndex PseudoObjectIndex
	 * @return Liefert den Wahrheitswert `true` zurück, falls bei der Berechnung des Arguments ein Fehler aufgetreten ist, sonst `false`.
	 */
	public ExpressionResult evaluate(
			List<ExpressionTree> argumentExpressions,
			Hashtable<RealElement, BaseDataSet> realElements,
			Hashtable<TempAttributeDescription, TempAttribut> tempElements, final int pseudoObjectIndex) {
		if(argumentExpressions == null || argumentExpressions.size() != 1) {
			return ExpressionResult.error(ErrorMessageBuilder.illegalArgumentCount(this, argumentExpressions, false, EnumSet.allOf(ExpressionResult.ResultType.class)));
		}
		else {
			ExpressionResult result = new ExpressionResult();
			ExpressionTree argumentExpression = argumentExpressions.get(0);
			ExpressionResultAndState argumentValue = argumentExpression.evaluateEx(realElements, tempElements, pseudoObjectIndex);
			result.set(getResult(argumentValue));
			return result;
		}
	}

	public ExpressionResult.ResultType getResultType(final List<ExpressionTree> argumentExpressions, final AST ast) throws SemanticErrorException {
		if(argumentExpressions == null || argumentExpressions.size() != 1) return ExpressionResult.ResultType.ERROR;
		ExpressionTree argumentExpression = argumentExpressions.get(0);
		// folgender Aufruf ist vermutlich notwendig, damit bestimmte Fehler erkannt werden
		argumentExpression.getResultType(ast);
		return ExpressionResult.ResultType.BOOL;
	}
}
