/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager.datamanager;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.NoncriticalException;
import de.bsvrz.pua.prot.interpreter.Interpreter;
import de.bsvrz.pua.prot.manager.ManagerErrorMessage;
import de.bsvrz.pua.prot.manager.requests.CreateScriptRequest;
import de.bsvrz.pua.prot.manager.requests.DataManagerRequest;
import de.bsvrz.pua.prot.manager.requests.ProtocolRequest;
import de.bsvrz.pua.prot.manager.requests.ScriptRequest;
import de.bsvrz.pua.prot.manager.status.StatePublisher;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.AtlMeta;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;

/**
 * Submodul Daten Manager.<br>
 * 1. Übernimmt die Verwaltung von Dateisystem-Operationen:<br>
 * - Schreiben von Protokollen<br>
 * - Herausfinden, ob eine bestimmte Protokoll-Id bereits vergeben wurde<br>
 * - Lesen und Schreiben von Skriptquelltexten<br>
 * 2. Zugriff auf die Konfiguration zur Verwaltung von Skript-Quelltexten<br>
 * 3. Zugriff auf den Skript-Interpreter <br>
 * <p>
 * Geht eine Anfrage über die Methoden {@link #addProtocolRequest(ResultData[])}
 * oder {@link #addScriptRequest(ResultData[])} ein, so wird diese Anfrage in
 * einem Ringpuffer abgelegt. Ist der Thread unbeschäftigt, werden sie der Reihe
 * nach (FIFO) abgearbeitet.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class DataManager extends Thread implements DataManagerInterface {

	/** Dateiendung Skript-Quelltext */
	private static final String SOURCE_FILENAME_EXTENSION = ".pua";

	/** Dateieindung Skript-Metainformationen */
	private static final String META_FILENAME_EXTENSION = ".mta";

	/** Initiale Ringbuffer Größe. */
	static private final int INITIAL_RINGBUFFER_SIZE = 1000;

	/**
	 * Ringpuffer in dem die Aufträge abgelegt werden, die bereit zur Abarbeitung
	 * sind
	 */
	private RingBuffer<DataManagerRequest> _requests;

	/** Datenmodell */
	private DataModel _model;

	/** Verbindung zum Datenverteiler */
	private ClientDavInterface _dav;

	/** Debug - Ausgaben */
	private static final Debug _debug = Debug.getLogger();

	/**
	 * Pfad unter dem die Dateien Verwaltet werden. Endet mit
	 * {@link File#separator}.
	 */
	private String _path;

	/** Minimaler verbleibender Speicherplatz auf dem Datenträgern (in MB) */
	private int _quota;

	/** Abbrechen des DatenManagers. */
	private boolean _abort = false;

	/** Auftrag der momentan abgearbeitet wird. */
	private DataManagerRequest _currentRequest;

	/** Steuert ob der Pretty Printer verwendet werden soll. */
	private boolean _prettyPrinter;

	/**
	 * Erzeugen des Data Managers. Nach dem Erzeugen des Data Manager Objekts muss
	 * der Data Manager mit {@link Thread#start()} gestartet werden.
	 *
	 * @param dav           Verbindung zum Datenverteiler
	 * @param path          Pfad unter dem die Dateien (Skriptquelltexte und
	 *                      Protokolldateien) verwaltet werden.
	 * @param quota         Minimaler verbleibender Speicherplatz auf dem
	 *                      Datenträgern (in MB)
	 * @param prettyPrinter Steuert ob der Pretty Printer verwendet werden soll.
	 *
	 * @throws FailureException Falls ein ungültiger Pfad übergeben wird.
	 */
	public DataManager(ClientDavInterface dav, String path, int quota, boolean prettyPrinter) throws FailureException {
		this._dav = dav;
		this._path = path;
		if (quota < 0)
			quota = 0;
		this._quota = quota;
		this._prettyPrinter = prettyPrinter;

		_requests = new RingBuffer<>(INITIAL_RINGBUFFER_SIZE, RingBuffer.UNBOUNDED_SIZE);
		_model = dav.getDataModel();

		if (path == null) {
			throw new FailureException(ManagerErrorMessage.INVALID_PATH, LoggerException.WARNING);
		}

		if (!path.endsWith(File.separator)) {
			path += File.separator;
		}
		File f = new File(path);
		if (!f.exists()) {
			throw new FailureException(ManagerErrorMessage.PATH_NONEXISTENT, LoggerException.WARNING);
		} else if (!f.isDirectory())
			throw new FailureException(ManagerErrorMessage.PATH_IS_NO_DIRECTORY, LoggerException.WARNING);
	}

	/**
	 * Startet die Interpretation eines Skriptes:<br>
	 * 1. Holt den Dateinamen des Skripts aus der Konfiguration<br>
	 * 2. Erzeugen eines Skript-Interpreters und evaluieren des Skripts<br>
	 * 3. Information zurückgeben<br>
	 *
	 * @param parameter Auftragsparameter
	 *
	 * @return Aus dem Skript gewonnene Informationen
	 *
	 * @throws FailureException Fehler beim Interpretieren des Skripts.
	 */
	public ProcessingInformation process(final ProcessingParameter parameter) throws FailureException {
		ProcessingInformation pi = null;
		DataInputStream dis = null;

		try {
			// Stream für Skriptquelltext öffnen
			String fileName = getAbsolutFileNameSource(parameter.getScript());
			dis = loadScriptSource(fileName);

			// Skriptinterpreter erzeugen und starten
			pi = new ProcessingInformation(parameter.getMainObject());
			Interpreter i = new Interpreter(_model);

			ProtocolRequestResult result = i.process(dis, fileName, parameter, pi);
			if (!result.isRequestSuccessful()) {
				throw new FailureException(result.getErrorMsg(), LoggerException.OTHER);
			} else {
				_debug.finest("Skriptinterpretation erfolgreich abgeschlossen");
			}
		} catch (FailureException e) {
			throw e;
		} finally {
			Util.close(dis);
		}

		return pi;
	}

	/**
	 * Öffnet Stream mit Skript Quelltext. Der Stream muss nach Gebrauch wieder
	 * geschlossen werden!
	 *
	 * @param fileName Dateiname
	 *
	 * @return InputStream der auf <code>fileName</code> zeigt.
	 *
	 * @throws FailureException Fehler beim Öffnen.
	 */
	private DataInputStream loadScriptSource(String fileName) throws FailureException {
		try {
			return new DataInputStream(new FileInputStream(fileName));
		}
		catch(FileNotFoundException e) {
			throw new FailureException(ManagerErrorMessage.INVALID_SCRIPT_FILENAME + fileName, e,
			                           LoggerException.WARNING
			);
		}	
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getMetaInformation(SystemObject)
	 */
	public AtlMeta getMetaInformation(SystemObject script) throws FailureException {
		AtlMeta meta = CreateScriptRequest.loadMetaInformation(getAbsolutFileNameMeta(script));
		// jetzt Skript einlesen und vergleichen:
		try {
			CreateScriptRequest.calcChecksum(readSource(script));
		} catch (NoSuchAlgorithmException e) {
			throw new FailureException(ManagerErrorMessage.INVALID_CHECKSUM_ALGORITHM
					+ CreateScriptRequest.CHECKSUM_ALGORITHM + ", " + e.getMessage(), LoggerException.WARNING);
		}
		return meta;
	}

	/**
	 * Liest den Quelltext eines Skripts ein.
	 *
	 * @return Skript Quelltext.
	 */
	private String readSource(SystemObject script) throws FailureException {
		String fileName = null;
		try {
			StringBuilder source = new StringBuilder();
			fileName = getAbsolutFileNameSource(script);
			try(BufferedReader in = Files.newBufferedReader(Paths.get(fileName), StandardCharsets.ISO_8859_1)) {
				String line;
				do {
					line = in.readLine();
					if(line != null)
						source.append(line).append("\n");
				}
				while(line != null);
			}
			return source.toString();
		} catch (FileNotFoundException e) {
			_debug.fine(ErrorMessage.FILE_NOT_FOUND + fileName + ", " + e.getMessage());
			throw new FailureException(ErrorMessage.FILE_NOT_FOUND + fileName + ", " + e.getMessage(), e,
					LoggerException.WARNING);
		} catch (IOException e) {
			_debug.fine(ErrorMessage.CAN_NOT_ACCESS_FILE + fileName + ", " + e.getMessage());
			throw new FailureException(ErrorMessage.CAN_NOT_ACCESS_FILE + fileName + ", " + e.getMessage(), e,
					LoggerException.WARNING);
		}
	}

	/**
	 * Arbeitet die Aufträge ab. Sind alle Aufträge erledigt, wartet der Thread, bis
	 * neue Aufträge eingehen. Vor beenden des Threads werden alle
	 * Schreiboperattionen abgebrochen.
	 */
	@Override
	public void run() {
		try {
			while (!_abort) {
				_currentRequest = _requests.pop();
				_currentRequest.process();
				_currentRequest = null;
				synchronized (_requests) {
					while (_requests.isEmpty() && !_abort)
						_requests.wait();
				}
			}
		} catch (InterruptedException e) {
			_debug.fine(ErrorMessage.INTERRUPTED_EXCEPTION + this.getClass().getName());
		}
	}

	/**
	 * Überprüft ob es zu der gegebenen Protokoll-Id bereits ein gespeichertes
	 * Protokoll gibt. Wenn nicht, dann ist die Protokoll-Id eindeutig.
	 *
	 * @param protocolId Protokoll-Id
	 *
	 * @return true falls die protokoll-Id bereits verwendet wurde und das Protokoll
	 *         auf dem Datenträger gespeichert wurde, false sonst.
	 */
	public boolean isUnique(long protocolId) {
		return !(new File(_path + protocolId + ProtocolOutputStream.PROTOCOL_FILENAME_EXTENSION).exists()
				|| new File(_path + protocolId + ProtocolOutputStream.PROTOCOL_FRAGMENT_EXTENSION).exists()
				|| new File(_path + protocolId + ProtocolOutputStream.PROTOCOL_NOT_READ_YET).exists());
	}

	/**
	 * Übergeben einer Skript-bezogenen Anfrage.
	 *
	 * @param resultData Objekt mit den Anfrageparametern
	 */
	public void addScriptRequest(ResultData resultData[]) {
		synchronized (_requests) {
			_requests.push(new ScriptRequest(_dav, this, resultData, _prettyPrinter));
			_requests.notifyAll();
		}
	}

	/** @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getPath() */
	public String getPath() {
		return _path;
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#process(java.lang.String)
	 */
	public ProcessingInformation process(String source) throws CriticalParserException {
		if (source == null || source.isEmpty())
			throw new CriticalParserException(ManagerErrorMessage.NO_SOURCE);
		ProcessingInformation pi = new ProcessingInformation();

		ByteArrayInputStream bis = new ByteArrayInputStream(source.getBytes(StandardCharsets.ISO_8859_1));
		DataInputStream dis = new DataInputStream(bis);

		Interpreter interpreter = new Interpreter(_model);
		ProtocolRequestResult result = interpreter.process(dis, "neu anzulegendem Skript", null, pi);

		try {
			dis.close();
		} catch (Exception e) {
			_debug.error("Fehler beim Schließen eines Streams: ", e);
		}

		if (!result.isRequestSuccessful())
			throw new CriticalParserException(result.getErrorMsg());

		return pi;
	}

	/**
	 * Übergeben einer Protokoll-bezogenen Anfrage.
	 *
	 * @param resultData Objekt mit den Anfrageparametern
	 */
	public void addProtocolRequest(ResultData resultData[]) {
		_requests.push(new ProtocolRequest(_dav, resultData, this));
		synchronized (_requests) {
			_requests.notifyAll();
		}
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getFileNameSource(SystemObject)
	 */
	public String getFileNameSource(SystemObject script) {
		return script.getId() + SOURCE_FILENAME_EXTENSION;
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getFileNameMeta(SystemObject)
	 */
	public String getFileNameMeta(SystemObject script) {
		return script.getId() + META_FILENAME_EXTENSION;
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getAbsolutFileNameMeta(SystemObject)
	 */
	public String getAbsolutFileNameMeta(SystemObject script) {
		return getPath() + getFileNameMeta(script);
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getAbsolutFileNameSource(SystemObject)
	 */
	public String getAbsolutFileNameSource(SystemObject script) {
		return getPath() + getFileNameSource(script);
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#addScriptRequest(de.bsvrz.pua.prot.manager.requests.ScriptRequest)
	 */
	public void addScriptRequest(ScriptRequest request) {
		synchronized (_requests) {
			_requests.push(request);
		}
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#shutdown()
	 */
	public void shutdown() {
		_abort = true; // -> dadurch wird der DataManager-Thread beendet, und alle Dateien geschlossen.
		Object notifier = _requests;
		synchronized (notifier) {
			_requests = new RingBuffer<>(RingBuffer.MIN_CHUNK_SIZE, RingBuffer.UNBOUNDED_SIZE);
			notifier.notifyAll();
		}
		if (_currentRequest != null)
			_currentRequest.abort();
		this.interrupt();
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#deleteProtocol(long)
	 */
	public void deleteProtocol(long protocolId) throws NoncriticalException {
		File savedFile = new File(_path + protocolId + ProtocolOutputStream.PROTOCOL_FILENAME_EXTENSION);
		File unreadFile = new File(_path + protocolId + ProtocolOutputStream.PROTOCOL_NOT_READ_YET);
		if (savedFile.exists()) {
			try {
				if (!savedFile.delete()) {
					throw new NoncriticalException(ManagerErrorMessage.CAN_NOT_DELETE_FILE + savedFile,
							LoggerException.WARNING);
				}
				StatePublisher.getInstance(_dav).notifyProtocolRemoved(protocolId);
			} catch (SecurityException e) {
				throw new NoncriticalException(ErrorMessage.CAN_NOT_ACCESS_FILE + protocolId, LoggerException.WARNING);
			}
		} else if (unreadFile.exists()) {
			try {
				if (!unreadFile.delete()) {
					throw new NoncriticalException(ManagerErrorMessage.CAN_NOT_DELETE_FILE + unreadFile,
							LoggerException.WARNING);
				}
				StatePublisher.getInstance(_dav).notifyProtocolRemoved(protocolId);
			} catch (SecurityException e) {
				throw new NoncriticalException(ErrorMessage.CAN_NOT_ACCESS_FILE + protocolId, LoggerException.WARNING);
			}
		} else {
			throw new NoncriticalException(ManagerErrorMessage.NO_PROTOCOL + protocolId, LoggerException.WARNING);
		}
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getSavedProtocol(long)
	 */
	public ProtocolInputStream getSavedProtocol(long protocolId) throws FailureException {
		return new ProtocolInputStream(protocolId, _path);
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getUnreadProtocol(long)
	 */
	public ProtocolInputStream getUnreadProtocol(long protocolId) throws FailureException {
		return new ProtocolInputStream(protocolId, _path);
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#createProtocolStream(long,de.bsvrz.pua.prot.util.ProcessingParameter)
	 */
	public ProtocolOutputStream createProtocolStream(long protocolId, ProcessingParameter pp)
			throws NoncriticalException {
		return new ProtocolOutputStream(protocolId, _path, pp, _quota);
	}

	/**
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface#getUpcomingOps(DataManagerRequest[])
	 */
	public DataManagerRequest[] getUpcomingOps(DataManagerRequest[] dmr) {
		return _requests.toArray(dmr);
	}
}
