/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager.requests;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.manager.ManagerErrorMessage;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.pua.prot.manager.datamanager.DataManager;
import de.bsvrz.pua.prot.manager.datamanager.ProtocolOutputStream;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.PidProtocol;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

/**
 * Klasse die anhand eines Objekts vom Typ {@link ResultData} überprüft, welcher Protokoll-bezogene Auftrag ausgeführt werden soll.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolRequest implements DataManagerRequest {

	/** Auftragsparameter */
	private ResultData _resultData[];

	/** Verbindung zum Datenverteiler */
	protected ClientDavInterface dav;

	/** Debug - Ausgabe */
	protected static final Debug debug = Debug.getLogger();

	/** Data Manager */
	protected DataManager dataManager;

	/** Zeigt an ob der Auftrag abgebrochen werden soll. */
	private boolean _abort;

	/**
	 * Erzeugt den Auftrag
	 *
	 * @param dav         Verbindung zum Datenverteiler
	 * @param resultData  Anfrageparameter
	 * @param dataManager Data Manager
	 */
	public ProtocolRequest(ClientDavInterface dav, ResultData resultData[], DataManager dataManager) {
		this.dav = dav;
		this._resultData = resultData;
		this.dataManager = dataManager;
	}

	/**
	 * Erzeugt konkrete Aufträge aus den Anfrageparametern und führt diese aus. Es werden folgende Anfragen unterstützt:<br> 1. Liste der gespeicherten Protokolle
	 * bzw. noch nicht abgerufenen Protokolle zurückgeben. Dazu wird ein Objekt vom Typ {@link ProtocolListRequest} erzeugt, das den Auftrag ausführt. 2. Löschen
	 * eines gespeicherten Protokolls: {@link DeleteProtocolRequest} führt den Auftrag aus.
	 *
	 * @return Liefert immer eine positive Rückmeldung
	 */
	public ProtocolRequestResult process() {
		for(ResultData rd : _resultData) {
			if(isAborted()) break;
			Data data = rd.getData();
			if(data != null) {
				int opCode = data.getUnscaledValue(PidProtocol.opCode).intValue();
				switch(opCode) {
					case ProtocolRelatedOperation.UNREAD_PROTOCOL_LIST:
						ProtocolListRequest upr = new ProtocolListRequest(dav, data, rd, dataManager, ProtocolOutputStream.PROTOCOL_NOT_READ_YET);
						upr.process();
						break;
					case ProtocolRelatedOperation.LIST:
						ProtocolListRequest rpr = new ProtocolListRequest(dav, data, rd, dataManager, ProtocolOutputStream.PROTOCOL_FILENAME_EXTENSION);
						rpr.process();
						break;
					case ProtocolRelatedOperation.DELETE:
						SystemObject client = data.getReferenceValue(PidProtocol.sender).getSystemObject();
						long requestId = data.getUnscaledValue(PidProtocol.protocolId).longValue();
						DeleteProtocolRequest dpr = new DeleteProtocolRequest(dav, rd, dataManager, client, requestId);
						dpr.process();
						break;
					default:
						SystemObject so = data.getReferenceValue(PidProtocol.sender).getSystemObject();
						if(so != null) {
							try {
								ProtocolAnswer.sendError(
										ProtocolAnswer.sender(dav, so), data.getUnscaledValue(PidProtocol.protocolId).longValue(), ManagerErrorMessage.INVALID_OPCODE + opCode
								);
							}
							catch(FailureException e) {
								debug.warning("Fehler beim Senden einer Fehlermeldung: ", e);
							}
						}
				}
			}
		}
		return new ProtocolRequestResult();
	}

	/** @see de.bsvrz.pua.prot.manager.requests.DataManagerRequest#abort() */
	public void abort() {
		this._abort = true;
	}

	/** @return Zeigt an ob der Auftrag abgebrochen werden soll. */
	public boolean isAborted() {
		return _abort;
	}
}
