/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager.status;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.pua.prot.manager.clientDatabase.ClientDatabase;
import de.bsvrz.pua.prot.manager.taskmanager.ThreadElement;
import de.bsvrz.pua.prot.sender.PuaSender;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.AtgProtocolRequest;
import de.bsvrz.sys.funclib.losb.datk.PidProtocol;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.util.Map;

/**
 * Veröffentlicht den Status der Protokolle über den Datenverteiler (derzeit nur Fertigmeldungen)
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
public class DavPublisher extends StatePublisherTask {

	private static final Debug _debug = Debug.getLogger();
	
	private static Object _lock = new Object();

	private final ClientDavInterface _davConnection;
	
	private static boolean _archiveAvailable = false;

	public DavPublisher(final ClientDavInterface clientDavInterface) {
		_davConnection = clientDavInterface;
	}

	/**
	 * Fügt einen neuen Client hinzu
	 *  @param dav DaV
	 * @param object               Empfänger-Objekt des Clients, an das Benachrichtigungen gesendet werden sollen
	 * @param protocolVersion      Protokollversion des Clients (bestimmt unter anderem ob erweiterte Benachrichtigungen an ihn gesendet werden, ab Version 2)
	 */
	public static void addClient(final ClientDavInterface dav, final SystemObject object, final int protocolVersion) {
		synchronized(_lock) {
			if(ClientDatabase.getClients().containsKey(object)) return; // bereits vorhanden
			PuaSender sender = new PuaSender(dav, object, getDataDescription(dav), () -> removeClient(object));
			ClientDatabase.addClient(object, sender, protocolVersion);
			if(protocolVersion >= 2) {
				try {
					final Data data = getNotifyArchiveData(dav, object, _archiveAvailable ? 1 : 0);
					sender.send(data);
				}
				catch(FailureException e) {
					_debug.warning("Kann Archivverfügbarkeit nicht versenden", e);
				}
			}
		}
	}

	private static Data getNotifyArchiveData(final ClientDavInterface dav, final SystemObject object, final int requestId) throws FailureException {
		return makeData(dav, requestId, object, new byte[]{3, ClientDatabase.CURRENT_PROTOCOL_VERSION});
	}

	public static void removeClient(final SystemObject object) {
		synchronized(_lock) {
			ClientDatabase.removeClient(object);
		}
	}

	private static DataDescription getDataDescription(final ClientDavInterface dav) {
		DataModel model = dav.getDataModel();
		AttributeGroup atg = model.getAttributeGroup(PidProtocol.atgProtocolRequest);
		Aspect asp = model.getAspect(PidProtocol.aspAnswer);
		return new DataDescription(atg, asp);
	}

	@Override
	public int getInterval() {
		return 60 * 1000;
	}

	@Override
	protected void update(final StatePublisher.Status status) {
	}

	@Override
	public void notifyThreadFinished(final ThreadElement threadElement) {
		try {
			synchronized(_lock) {
				for(final ClientDatabase.Client client : ClientDatabase.getClients().values()) {
					byte[] dataBytes;
					dataBytes = serializeThreadInfo(threadElement);
					final Data data = makeData(_davConnection, threadElement.getJobId(), client.getObject(), dataBytes);
					client.getClientSender().send(data);
				}
			}
		}
		catch(FailureException e) {
			_debug.warning("Konnte Protokoll-Fertigmeldung nicht verschicken", e);
		}
	}

	private byte[] serializeThreadInfo(final ThreadElement threadElement) {
		try {
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			out.write(1);
			ObjectOutputStream stream = new ObjectOutputStream(out);
			stream.writeObject(threadElement.toJobInProgress());
			stream.close();
			return out.toByteArray();
		}
		catch(Exception e) {
			_debug.warning("Konnte Protokollauftrag nicht serialisieren", e);
			return new byte[]{1};
		}
	}

	@Override
	public void notifyProtocolRemoved(final long protocolId) {
		try {
			synchronized(_lock) {
				for(final ClientDatabase.Client client : ClientDatabase.getClients().values()) {
					if(client.getProtocolVersion() >= 2) {
						final Data data = makeData(_davConnection, protocolId, client.getObject(), new byte[]{ 0 });
						client.getClientSender().send(data);
					}
				}
			}
		}
		catch(FailureException e) {
			_debug.warning("Konnte Protokoll-Löschen-Meldung nicht verschicken", e);
		}
	}

	@Override
	public void notifyProtocolRead(final long protocolId) {
		try {
			synchronized(_lock) {
				for(final ClientDatabase.Client client : ClientDatabase.getClients().values()) {
					if(client.getProtocolVersion() >= 2) {
						final Data data = makeData(_davConnection, protocolId, client.getObject(), new byte[]{2});
						client.getClientSender().send(data);
					}
				}
			}
		}
		catch(FailureException e) {
			_debug.warning("Konnte Protokoll-Gelesen-Meldung nicht verschicken", e);
		}
	}

	public void notifyArchiveAvailable(final boolean archiveAvailable) {
		_archiveAvailable = archiveAvailable;
		try {
			synchronized(_lock) {
				for(final ClientDatabase.Client client : ClientDatabase.getClients().values()) {
					if(client.getProtocolVersion() >= 2) { 
						// Eigentlich >= 3, aber man kann hier nicht einwandfrei zwischen Clientversion 2 und 3 unterscheiden und einen Client in Version 2 stört die Nachricht nicht
						final Data data = getNotifyArchiveData(_davConnection, client.getObject(), archiveAvailable ? 1 : 0);
						client.getClientSender().send(data);
					}
				}
			}
		}
		catch(FailureException e) {
			_debug.warning("Konnte Protokoll-Gelesen-Meldung nicht verschicken", e);
		}
	}

	private static Data makeData(final ClientDavInterface davConnection, final long requestId, final SystemObject client, final byte[] data) throws FailureException {
		return AtgProtocolRequest.build(davConnection, client, requestId, ProtocolRelatedOperation.PROTOCOL_COMPLETION_INTERFACE, data);
	}
}
