/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.data;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.IntegerAttributeType;
import de.bsvrz.dav.daf.main.config.IntegerValueRange;
import de.bsvrz.dav.daf.main.config.IntegerValueState;
import de.bsvrz.pua.prot.util.ExpressionResult;

import java.text.NumberFormat;
import java.util.Collections;
import java.util.List;
import java.util.Locale;

/**
 * Unskalierte Ansicht auf ein Datenobjekt.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
class ProtocolDataUnscaledValue extends ProtocolDataTextValue implements Data.NumberValue {

	protected final NumberFormat _doubleNumberFormat;

	/**
	 * Erstellt eine neue unskalierte Zahl-Ansicht auf ein Datum
	 *
	 * @param attributeProtocolData datum
	 */
	public ProtocolDataUnscaledValue(final AttributeProtocolData attributeProtocolData) {
		super(attributeProtocolData);

		_doubleNumberFormat = NumberFormat.getNumberInstance(Locale.GERMANY);
		_doubleNumberFormat.setMinimumIntegerDigits(1);
		_doubleNumberFormat.setMaximumIntegerDigits(999);
		final int digits = (int)Math.ceil(-Math.log10(getScaling()));
		_doubleNumberFormat.setMinimumFractionDigits(digits);
		_doubleNumberFormat.setMaximumFractionDigits(digits + 3);
		_doubleNumberFormat.setGroupingUsed(false);
	}

	public IntegerValueState getState() {
		try {
			final long myValue = getUnscaledLongValue();
			final List<IntegerValueState> valueStates = getStates();
			for(final IntegerValueState state : valueStates) {
				if(state.getValue() == myValue) return state;
			}
			return null;
		}
		catch(ArithmeticException ignored){
			// Kein Zahlenwert
			return null;
		}
	}

	public boolean isNumber() {
		if(_data.getAttributeType() instanceof IntegerAttributeType) {
			final IntegerAttributeType integerAttributeType = (IntegerAttributeType)_data.getAttributeType();
			final IntegerValueRange range = integerAttributeType.getRange();
			if(range == null) return false;
			try {
				final long value = getUnscaledLongValue();
				if(value < range.getMinimum() || value > range.getMaximum()) return false;
			}
			catch(ArithmeticException ignored) {
				// Wenn das Datum keine Zahl darstellt, sollte diese Funktion false zurückliefern
				return false;
			}
		}
		return true;
	}

	public boolean isState() {
		return getState() == null;
	}

	private long boolToInt() {
		final boolean boolValue = _data.getInternalValue().getBool();
		final List<IntegerValueState> valueStates = getStates();
		for(final IntegerValueState state : valueStates) {
			if(boolValue && "Ja".equals(state.getName())) {
				return state.getValue();
			}
			if(!boolValue && "Nein".equals(state.getName())) {
				return state.getValue();
			}
		}
		throw new UnsupportedOperationException("Ungültiger AttributTyp für Boolean-Werte: " + _data.getAttributeType());
	}

	/**
	 * Gibt eine Liste mit Zustandswerten zurück oder eine leere Liste wenn der Attributtyp dies nicht unterstützt.
	 *
	 * @return eine Liste mit Zustandswerten zurück oder eine leere Liste
	 */
	private List<IntegerValueState> getStates() {
		if(_data.getAttributeType() instanceof IntegerAttributeType) {
			final IntegerAttributeType integerAttributeType = (IntegerAttributeType)_data.getAttributeType();
			return integerAttributeType.getStates();
		}
		return Collections.emptyList();
	}

	public byte byteValue() {
		return (byte)longValue();
	}

	public short shortValue() {
		return (short)longValue();
	}

	public int intValue() {
		return (int)longValue();
	}

	public long longValue() {
		return getUnscaledLongValue();
	}

	public float floatValue() {
		return (float)doubleValue();
	}

	public double doubleValue() {
		return getUnscaledDoubleValue();
	}

	/**
	 * Gibt den unskalierten Long-Wert dieses Datums zurück
	 *
	 * @return den unskalierten Long-Wert dieses Datums
	 *
	 * @throws ArithmeticException Datum hat keinen Zahlenwert
	 */
	final long getUnscaledLongValue() {
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.DOUBLE) {
			return Math.round(getReverseScaledValue());
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.LONG) {
			return _data.getInternalValue().getLong();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.BOOL) {
			return boolToInt();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.STATUS) {
			final String stateName = _data.getInternalValue().getString();
			for(final IntegerValueState state : getStates()) {
				if(state.getName().equals(stateName)) {
					return state.getValue();
				}
			}
		}
		throw new ArithmeticException("Das Datum hat keinen Zahlenwert");
	}

	/**
	 * Gibt den unskalierten Double-Wert dieses Datums zurück
	 *
	 * @return den unskalierten Double-Wert dieses Datums
	 *
	 * @throws ArithmeticException Datum hat keinen Zahlenwert
	 */
	final double getUnscaledDoubleValue() {
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.DOUBLE) {
			return getReverseScaledValue();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.LONG) {
			return _data.getInternalValue().getLong();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.BOOL) {
			return boolToInt();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.STATUS) {
			final String stateName = _data.getInternalValue().getString();
			for(final IntegerValueState state : getStates()) {
				if(state.getName().equals(stateName)) {
					return state.getValue();
				}
			}
		}
		throw new ArithmeticException("Das Datum hat keinen Zahlenwert");
	}

	/**
	 * Gibt den skalierten Long-Wert dieses Datums zurück
	 *
	 * @return den skalierten Long-Wert dieses Datums
	 *
	 * @throws ArithmeticException Datum hat keinen Zahlenwert
	 */
	final long getScaledLongValue() {
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.DOUBLE) {
			return Math.round(_data.getInternalValue().getDouble());
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.LONG) {
			final double scaling = getScaling();
			if(scaling == 1.0) return _data.getInternalValue().getLong();
			return Math.round(_data.getInternalValue().getLong() * getScaling());
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.BOOL) {
			return boolToInt();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.STATUS) {
			final String stateName = _data.getInternalValue().getString();
			for(final IntegerValueState state : getStates()) {
				if(state.getName().equals(stateName)) {
					return Math.round(state.getValue() * getScaling());
				}
			}
		}
		throw new ArithmeticException("Das Datum hat keinen Zahlenwert");
	}

	/**
	 * Gibt den skalierten Double-Wert dieses Datums zurück
	 *
	 * @return den skalierten Double-Wert dieses Datums
	 *
	 * @throws ArithmeticException Datum hat keinen Zahlenwert
	 */
	final double getScaledDoubleValue() {
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.DOUBLE) {
			return _data.getInternalValue().getDouble();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.LONG) {
			return _data.getInternalValue().getLong() * getScaling();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.BOOL) {
			return boolToInt();
		}
		if(_data.getInternalValue().getType() == ExpressionResult.ResultType.STATUS) {
			final String stateName = _data.getInternalValue().getString();
			for(final IntegerValueState state : getStates()) {
				if(state.getName().equals(stateName)) {
					return state.getValue() * getScaling();
				}
			}
		}
		throw new ArithmeticException("Das Datum hat keinen Zahlenwert");
	}

	private double getReverseScaledValue() {
		return _data.getInternalValue().getDouble() / getScaling();
	}

	/**
	 * Gibt den Skalierungsfaktor zurück
	 *
	 * @return Skalierungsfaktor wie in {@link de.bsvrz.dav.daf.main.config.IntegerValueRange#getConversionFactor()} definiert, oder 1.0 falls unskaliert
	 */
	final double getScaling() {
		if(_data.getAttributeType() instanceof IntegerAttributeType) {
			final IntegerAttributeType integerAttributeType = (IntegerAttributeType)_data.getAttributeType();
			final IntegerValueRange range = integerAttributeType.getRange();
			if(range == null) return 1.0;
			return range.getConversionFactor();
		}
		return 1.0;
	}

	@Override
	public String getValueText() {
		try {
			final IntegerValueState valueState = getState();
			if(valueState != null) return valueState.getName();
			if(_data.getInternalValue().getType() == ExpressionResult.ResultType.LONG) return Long.toString(longValue());
			return _doubleNumberFormat.format(doubleValue());
		}
		catch(NumberFormatException ignored) {
			return _data.getInternalValue().toString();
		}
	}

	@Override
	public String getSuffixText() {
		try {
			final IntegerValueState valueState = getState();
			if(valueState != null) return "";
			if(_data.getAttributeType() instanceof IntegerAttributeType) {
				final IntegerAttributeType integerAttributeType = (IntegerAttributeType)_data.getAttributeType();
				if(integerAttributeType == null) return "";
				final IntegerValueRange range = integerAttributeType.getRange();
				if(range == null) return "";
				return range.getUnit();
			}
			return "";
		}
		catch(NumberFormatException ignored) {
			return "";
		}
	}

	public void setState(final IntegerValueState state) {
		throw new UnsupportedOperationException(ProtocolData.MESSAGE_NO_CHANGE);
	}

	public void set(final int value) {
		throw new UnsupportedOperationException(ProtocolData.MESSAGE_NO_CHANGE);
	}

	public void set(final long value) {
		throw new UnsupportedOperationException(ProtocolData.MESSAGE_NO_CHANGE);
	}

	public void set(final float value) {
		throw new UnsupportedOperationException(ProtocolData.MESSAGE_NO_CHANGE);
	}

	public void set(final double value) {
		throw new UnsupportedOperationException(ProtocolData.MESSAGE_NO_CHANGE);
	}
}
