/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import antlr.collections.AST;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.dav.daf.main.config.DoubleAttributeType;
import de.bsvrz.dav.daf.main.config.IntegerAttributeType;
import de.bsvrz.dav.daf.main.config.StringAttributeType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.TimeAttributeType;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ExpressionTree;
import de.bsvrz.pua.prot.util.MemberCheck;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.attributes.AttributeDescription;
import de.bsvrz.pua.prot.util.attributes.Elements;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;
import de.bsvrz.sys.funclib.losb.datk.PidMisc;

import java.util.*;

/**
 * Semantische Überprüfung der temporären Attribute.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class TempAttributeSemantics implements SemanticChecks {

	private Semantics _semantics;

	private SymbolTable _symbolTable;

	private AttributeSemantics _attributeSemantics;

	private MemberCheck _check;

	/** Liste mit allen temporären Attributen. */
	private Hashtable<String, TempAttributeDescription> _allTempAttributes;


	/**
	 * @param semantics Semantics
	 * @param attributeSemantics AttributeSemantics
	 */
	public TempAttributeSemantics(final Semantics semantics, final AttributeSemantics attributeSemantics) {
		this._semantics = semantics;
		this._check = semantics.getCheck();
		this._symbolTable = semantics.symbolTable;
		this._attributeSemantics = attributeSemantics;

		_allTempAttributes = new Hashtable<>();
	}

	/**
	 * Überprüft die Definition eines temporären Attributs. Falls es auf ein reales Attribut verweist, wird der Attribut-Zähler {@link
	 * AttributeSemantics#incAttributeCount()} erhöht.
	 *
	 * @param ast        Knoten (wird nur für Zeilenangabe in der Fehlermeldung benötigt)
	 * @param attribute  Temporäres Attribut
	 * @param type       Typ des temporären Attributs
	 * @param expression Ausdruck um den Wert des temporären AttributeDescription zu berechnen
	 *
	 * @throws CriticalParserException Falls das Attribut nicht korrekt definiert wurde oder es zu einem Kommunikationsfehler kommt
	 */
	public void addTempAttributeDefinition(final AST ast, final String attribute, final String type, final ExpressionTree expression) throws CriticalParserException {
		if(expression == null) throw new SemanticErrorException(InterpreterErrorMessage.EXPRESSION_EXPECTED);
		try {
			if(_allTempAttributes.containsKey(attribute))	//prüfen ob Name eindeutig
			{
				throw new CriticalParserException(InterpreterErrorMessage.REDEFINITION_TEMP_ATTRIBUTE + attribute);
			}
			else {
				final List<AttributeDescription> attSet = expression.getAttributes();

				for(final AttributeDescription att : attSet) {
					if(att instanceof RealAttributeDescription) {
						_attributeSemantics.incAttributeCount();
						break;
					}
				}

				final SystemObject objectType = _check.getObject(type);
				if(objectType == null) throw new SemanticErrorException(InterpreterErrorMessage.INVALID_PID + type);

				TempAttributeDescription tempAttributeDescription = _symbolTable.getOrCreateTempAttributeDefinition(attribute);

				tempAttributeDescription.setAST(ast);
				tempAttributeDescription.setExpression(expression);
				tempAttributeDescription.setType(objectType);

				_allTempAttributes.put(tempAttributeDescription.getName(), tempAttributeDescription);

			}
		}
		catch(CriticalParserException criticalException) {
			criticalException.line = ast.getLine();
			throw criticalException;
		}
	}

	/**
	 * Gibt an, ob der Attributtyp ein sinnvoller Container für den Wert einer Berechnung (z.B. in TempAttributen) ist. Folgende Konvertierungen werden zur Zeit erlaubt:<ul>
	 * <li>bool -> att.JaNein, StringAttributeType</li>
	 * <li>string -> StringAttributeType</li>
	 * <li>long, double -> DoubleAttributeType, IntegerAttributeType, TimeAttributeType, StringAttributeType</li>
	 * </ul>
	 * @param attributeType Attributtyp in den der Wert gesteckt werden soll
	 * @param resultType Ergebnistyp einer Berechnung
	 * @return true wenn Zuweisung erlaubt
	 * @throws SemanticErrorException Wenn ein Integer-Attribut keinen Zahlenbereich enthält, es aber Zahlen aufnehmen soll
	 */
	static boolean isTypeCompatible(final AttributeType attributeType, final ExpressionResult.ResultType resultType) throws SemanticErrorException {

		final boolean isDoubleAttribute = attributeType instanceof DoubleAttributeType;
		final boolean isTimeAttribute = attributeType instanceof TimeAttributeType;
		final boolean isIntegerAttribute = attributeType instanceof IntegerAttributeType;
		final boolean isStringAttribute = attributeType instanceof StringAttributeType;

		switch(resultType) {
			case BOOL:
				return attributeType.getPid().equals(PidMisc.PID_BOOLEAN_TYPE) || isStringAttribute;
			case STRING:
				return isStringAttribute;
			case LONG:
			case DOUBLE:
				if(isIntegerAttribute) {
					final IntegerAttributeType integerAttributeType = (IntegerAttributeType)attributeType;
					if(integerAttributeType.getRange() == null) {
						throw new SemanticErrorException(
								InterpreterErrorMessage.INVALID_TEMP_ATTRIBUTE + attributeType.toString() + " (Rückgabe des Ausdrucks: " + resultType + ")"
						);
					}
				}
				return (isIntegerAttribute || isTimeAttribute || isDoubleAttribute || isStringAttribute);
		}
		return false;
	}

	/**
	 * Fügt ein Platzhalter Attribut in die Liste der Attribute bzw. Spalten ein. Dies ist notwendig, damit die Reihenfolge der Attribute bestehen bleibt! Der Platzhalter muss
	 * abschließend (@see #cleanUp(ProcessingParameter, boolean)} durch das jeweilige temporäre Attribut ersetzt werden.
	 *
	 * @param name Name des temporären Attributs
	 * @param columnName ColumnName
	 */
	public void addTempAttributePlaceholder(final String name, final String columnName) {
		final TempAttributeDescription tad = new TempAttributeDescription(name, columnName);
		_semantics.pi.addElement(tad);
	}


	/**
	 * Führt eine abschließende Überprüfung durch.<br> - Stellt sicher, dass jedes verwendete temporäre Attribut eine gültige Definition besitzt. <br> - Stellt
	 * sicher, dass es keine Rekursion aufgetreten ist.<br> - Ersetzt Platzhalter-Attribute(von {@link #addTempAttributePlaceholder(String,String)} eingefügt) durch das jeweilige
	 * temporäre Attribut.
	 *
	 * @param pp            Wird nicht berücksichtigt.
	 * @param processScript Hat keine Bedeutung.
	 *
	 * @throws SemanticErrorException Undefiniertes temporäres Attribut.
	 */
	public void cleanUp(final ProcessingParameter pp, final boolean processScript) throws SemanticErrorException {
//		 Jedes temporäre Attribut wurde auch wirklich definiert
		if(_allTempAttributes.size() != _symbolTable.getTempAttributes().size()) {
			final StringBuffer errorMsg = new StringBuffer();
			errorMsg.append(InterpreterErrorMessage.UNDEFINED_TEMPATT);
			for(final TempAttributeDescription tempAttribute : _symbolTable.getTempAttributes()) {
				if(!_allTempAttributes.contains(tempAttribute.getName())) errorMsg.append(", " + tempAttribute);
			}
			throw new SemanticErrorException(errorMsg.toString());
		}

		final String errorMsg = checkForRecursion();
		if(errorMsg != null) throw new SemanticErrorException(errorMsg);

		//OPT kann ineffizient sein, wenn #reale Att >> #temp Att
		final List<Elements> elements = _semantics.pi.getAttributes();
		int i = 0;

		while(i < elements.size()) {
			final Elements element = elements.get(i);
			if(element instanceof TempAttributeDescription) {
				final TempAttributeDescription placeHolder = (TempAttributeDescription)element;

				final TempAttributeDescription realElement = _allTempAttributes.get(placeHolder.getName());

				// Hier geht es darum, herauszufinden, wieviele Spalten angelegt werden sollen, wenn tempattribute Alias definitionen verwenden.
				final int numColumns = realElement.getExpression().getNumColumns();

				final List<TempAttributeDescription> tad;

				if(numColumns == 1){
					tad = Arrays.asList(realElement.createIndexedClone(0));
				}
				else{
					final List<TempAttributeDescription> list = new ArrayList<>();
					for(int columnIndex = 0; columnIndex < numColumns; columnIndex++){
						list.add(realElement.createIndexedClone(columnIndex));
					}
					tad = list;
				}

				if(tad == null) {
					throw new SemanticErrorException(InterpreterErrorMessage.UNDEFINED_TEMPATT + placeHolder.getName());
				}
				else {
					_semantics.pi.replaceElement(i, tad, placeHolder.getColumnName());
					i += tad.size();
				}
			}
			else {
				i++;
			}
		}

		// Prüfen, ob das angegebene Attribut zum Ausdruck passt
		for(final TempAttributeDescription tempAttributeDescription : _allTempAttributes.values()) {
			// Den Rückgabetyp des Ausdrucks prüfen
			final ExpressionTree expression = tempAttributeDescription.getExpression();

			final ExpressionResult.ResultType resultType = expression.getResultType(tempAttributeDescription.getAST());

			if(tempAttributeDescription.getType() instanceof AttributeType) {
				final AttributeType attributeType = (AttributeType)tempAttributeDescription.getType();
				if(!isTypeCompatible(attributeType, resultType)) {
					throw new SemanticErrorException(
							InterpreterErrorMessage.INVALID_ATTRIBUTE_DEFINITION + tempAttributeDescription.getType() + " (Rückgabe des Ausdrucks: " + resultType + ")",
					        tempAttributeDescription.getAST()
					);
				}
			}
			else {
				throw new SemanticErrorException(InterpreterErrorMessage.INVALID_ATTRIBUTETYPE_PID + tempAttributeDescription.getType());
			}
		}
	}

	/**
	 * Überprüft ob die Definitionen der temporären Attribute rekursiv sind.
	 *
	 * @return Fehlermeldung, falls eine rekursive Definition gefunden wurde. Null sonst.
	 *
	 * @see #checkAttributeForRecursion(TempAttributeDescription,HashSet,HashSet)
	 */
	private String checkForRecursion() {
		//bereits geprüfte Attribute
		final HashSet<TempAttributeDescription> checked = new HashSet<>();
		//Attribute die sich gerade im Test befinden
		final HashSet<TempAttributeDescription> inTest = new HashSet<>();

		for(final TempAttributeDescription tad : _allTempAttributes.values()) {
				String result = checkAttributeForRecursion(tad, checked, inTest);
				if(result != null) return result;
			}
		return null;
	}

	/**
	 * Überprüft ob die Definitionen eines temporären Attributs rekursiv ist. Diese Methode ruft sich selbst rekursiv für jedes temporäre Attribut, auf das
	 * <code>startAttribute</code> verweist, auf.
	 *
	 * @param startAttribute Temporäres Attribut das überprüft werden soll.
	 * @param checked        Bisher überprüfte temporäre Attribute.
	 * @param inTest         Attribute die gerade überprüft werden.
	 *
	 * @return Fehlermeldung, falls eine rekursive Definition gefunden wurde. Null sonst.
	 */
	private String checkAttributeForRecursion(
			final TempAttributeDescription startAttribute, final HashSet<TempAttributeDescription> checked, final HashSet<TempAttributeDescription> inTest) {
		String result = null;
		if(inTest.contains(startAttribute))		//Von diesem Attribut ging der Test aus -> Rekursion
		{
			result = InterpreterErrorMessage.RECURSIVE_EXPRESSION + startAttribute;
		}
		else if(!checked.contains(startAttribute))		//der dem Attribut zugeordnete Ausdruck wurde noch nicht überprüft
		{
			inTest.add(startAttribute);
			for(final TempAttributeDescription tad : startAttribute.getTempAttributes()) {
				result = checkAttributeForRecursion(tad, checked, inTest);
				if(result != null) break;
			}
			checked.add(startAttribute);
			inTest.remove(startAttribute);
		}
		return result;
	}

	/** Diese Methode existiert nur zu Testzwecken und überschreibt {@link #_allTempAttributes}! */
	String checkForRecursion(final List<TempAttributeDescription> atts) {
		final Hashtable<String, TempAttributeDescription> table = new Hashtable<>();
		for(final TempAttributeDescription tad : atts) {
			table.put(tad.getName(), tad);
		}

		_allTempAttributes = table;
		return checkForRecursion();
	}
}
