/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager.datamanager;

import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.NoncriticalException;
import de.bsvrz.pua.prot.manager.requests.DataManagerRequest;
import de.bsvrz.pua.prot.manager.requests.ScriptRequest;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.sys.funclib.losb.datk.AtlMeta;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

import java.io.File;

/**
 * Schnittstelle des Submoduls Daten Manager. Der Daten Manager übernimmt die Verwaltung der Skriptquelltexte, Skriptobjekte und gespeicherten Protokolle.
 * Außerdem bietet Das Objekt, das diese Schnittstelle implementiert muss das Verzeichnis, in dem Skripte und Protokolle abgelegt werden kennen! Dies muss ggf.
 * über den Konstruktor eingestellt werden können.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public interface DataManagerInterface {

	/**
	 * Übergeben einer Skript-bezogenen Anfrage
	 *
	 * @param resultData Objekt mit den Anfrageparametern
	 */
	public void addScriptRequest(ResultData resultData[]);

	/**
	 * Übergeben einer Skript-bezogenen Anfrage
	 *
	 * @param request Anfrage
	 */
	public void addScriptRequest(ScriptRequest request);

	/**
	 * Übergeben einer Protokoll-bezogenen Anfrage
	 *
	 * @param resultData Objekt mit den Anfrageparametern
	 */
	public void addProtocolRequest(ResultData resultData[]);

	/**
	 * Startet die Interpretation eines Skriptes.
	 *
	 * @param parameter Auftragsparameter
	 *
	 * @return Aus dem Skript gewonnene Informationen.
	 *
	 * @throws FailureException Fehler beim Interpretieren des Skripts.
	 */
	public ProcessingInformation process(ProcessingParameter parameter) throws FailureException;

	/**
	 * Startet die Interpretation eines Skriptes. Dabei werden dem Interpreter keine Aufrufparameter übergeben. Daher können die Daten, die zurückgegeben werden,
	 * nicht von der Datenaufbereitung verwendet werden.
	 *
	 * @param source Source des Skriptobjekts
	 *
	 * @return Aus dem Skript gewonnene Informationen
	 *
	 * @throws CriticalParserException Fehlermeldung
	 */
	public ProcessingInformation process(String source) throws CriticalParserException;

	/**
	 * Überprüft ob es zu der gegebenen Protokoll-Id bereits ein gespeichertes Protokoll gibt. Wenn nicht, dann ist die Protokoll-Id eindeutig.
	 *
	 * @param protocolId Protokoll-Id
	 *
	 * @return true falls die protokoll-Id bereits verwendet wurde und das Protokoll auf dem Datenträger gespeichert wurde, false sonst.
	 */
	public boolean isUnique(long protocolId);

	/**
	 * Gibt den Pfad der Protokoll- und Skriptdateien zurück. Endet mit {@link File#separator}
	 *
	 * @return Pfad der Protokoll- und Skriptdateien.
	 */
	public String getPath();

	/**
	 * Liefert den Dateinamen der Metainformationen.
	 *
	 * @param script Skriptobjekt
	 *
	 * @return Dateiname
	 */
	public String getFileNameMeta(SystemObject script);

	/**
	 * Liefert Pfad und Dateiname der Metainformation
	 *
	 * @param script Skriptobjekt
	 *
	 * @return Pfad und Dateiname
	 */
	public String getAbsolutFileNameMeta(SystemObject script);

	/**
	 * Liefert den Dateinamen des Skriptquelltexts.
	 *
	 * @param script Skriptobjekt
	 *
	 * @return Dateiname
	 */
	public String getFileNameSource(SystemObject script);

	/**
	 * Liefert Pfad und Dateiname des Skriptquelltextes
	 *
	 * @param script Skriptobjekt
	 *
	 * @return Pfad und Dateiname
	 */
	public String getAbsolutFileNameSource(SystemObject script);

	/**
	 * Liefert die zum Skript gehörenden Metainformationen. Dazu werden die auf der Festplatte gespeicherten Informationen eingelesen. Überprüft zudem die
	 * Prüfsumme.
	 *
	 * @param script Skriptobjekt.
	 *
	 * @return Metainformationen oder <code>null</code> falls keine Metainformation existiert.
	 *
	 * @throws FailureException Korrupte Metainformation.
	 */
	public AtlMeta getMetaInformation(SystemObject script) throws FailureException;

	/** Fährt den Data-Manager herunter. */
	public void shutdown();

	/**
	 * Löscht ein gespeichertes Protokoll.
	 *
	 * @param protocolId ProtocolId
	 *
	 * @throws NoncriticalException Fehler beim Löschen.
	 */
	public void deleteProtocol(long protocolId) throws NoncriticalException;

	/**
	 * Liefert ein Objekt zum Zugriff auf ein gespeichertes Protokoll zurück.
	 *
	 * @param protocolId Id des Protokolls, das gelesen werden soll.
	 *
	 * @return Objekt zum Zugriff auf das gespeicherte Protokoll
	 *
	 * @throws FailureException Fehler.
	 */
	public ProtocolInputStream getSavedProtocol(long protocolId) throws FailureException;

	/**
	 * Liefert ein Objekt zum Zugriff auf ein gespeichertes Protokoll zurück.
	 *
	 * @param protocolId Id des Protokolls, das gelesen werden soll.
	 *
	 * @return Objekt zum Zugriff auf das gespeicherte Protokoll.
	 *
	 * @throws FailureException Fehler.
	 */
	public ProtocolInputStream getUnreadProtocol(long protocolId) throws FailureException;

	/**
	 * Erzeugt einen Ausgabestream für ein Protokoll.
	 *
	 * @param protocolId Protokoll-Id.
	 * @param pp         Auftragsparameter.
	 *
	 * @return Stream zum schreiben des Protokolls.
	 *
	 * @throws NoncriticalException Die Datei kann nicht angelegt werden; Fehler beim Schreiben der Protokollinformationen.
	 */
	public ProtocolOutputStream createProtocolStream(long protocolId, ProcessingParameter pp) throws NoncriticalException;

	/**
	 * Liefert eine Liste mit anstehenden Aufgaben.
	 *
	 * @param dmr Array, in den das Ergebnis kopiert wird. Ist der Array zu klein, so wird ein neuer Array angelegt.
	 *
	 * @return Anstehende Aufgaben.
	 */
	public DataManagerRequest[] getUpcomingOps(DataManagerRequest[] dmr);
}
