/*
 * Copyright 2019 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.onlinebuffer;

import de.bsvrz.dav.daf.accessControl.UserInfo;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.util.ReferenceQualifier;
import de.bsvrz.pua.prot.util.attributes.AttributeIdentifier;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;

/**
 * Objekt zum Empfang von Online-Daten für Referenzattribute. Gehen Daten ein, so wird sofort die Referenzkette durchlaufen, um die endgültigen Daten zu finden
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 */
class OnlineDataReferenceReceiverObject extends OnlineDataReceiverObject {

	/** Qualifizierung der Referenz */
	private final ReferenceQualifier qualifier;

	/** Zeigt an auf welche Daten angemeldet werden soll. */
	private final ReceiveOptions receiveOptions;

	/** Thread der die Referenzdaten einholt */
	private final GetRefThread collectorThread;

	/**
	 * @param receiveOptions   Empfangsoptionen
	 * @param receiverSet      Empfängergruppe, zu der dieser Empfänger gehört.
	 * @param qualifier        Datenidentifikation auf die sich der Receiver anmelden soll.
	 * @param newDataCollector Objekt das benachrichtig wird, wenn neue Daten vorliegen
	 * @param connection       Datenverteilerverbindung
	 * @param userInfo         Benutzer, dessen Rechte geprüft werden sollen
	 */
	public OnlineDataReferenceReceiverObject(
			ReceiveOptions receiveOptions,
			OnlineDataReceiverSet receiverSet,
			ReferenceQualifier qualifier,
			NewDataCollector newDataCollector,
			final ClientDavInterface connection,
			final UserInfo userInfo) {
		super(receiverSet, qualifier, newDataCollector, connection, userInfo);
		this.receiveOptions = receiveOptions;
		this.qualifier = qualifier;
		collectorThread = new GetRefThread(this);
		collectorThread.start();
	}

	/**
	 * Speichert den eingegangenen Datensatz. Da es sich um einen Empfänger für Referenzdaten handelt, wird die Referenzkette durchlaufen, und es wird der Wert der
	 * Referenz gespeichert. Da diese Methode vom der Dav-Callback-Methode {@link OnlineDataReceiverObject#update(ResultData[])} aufgerufen wird, muss ein neuer
	 * Thread gestartet werden, wenn eine Referenz abgeholt wird - sonst meldet der Dav keine weiteren Daten mehr.
	 *
	 * @see OnlineDataReceiverObject#storeData(ResultData)
	 */
	protected void getData(final ResultData result) {
		collectorThread.add(result);
	}

	/**
	 * Holt die Referenz vom Datenverteiler
	 *
	 * @param reference Datenidentifikation, die benötigt wird. (Das Objekt reference.object) wird nicht verwendet!
	 * @param object    Objekt, für das die Daten abgerufen werden sollen.
	 * @param dataTime  Zeitstempel des ursprünglichen Datensatzes
	 *
	 * @return Wert der Referenz
	 *
	 * @throws FailureException Fehler beim Datenempfang
	 */
	private ResultData takeReference(ReferenceQualifier reference, SystemObject object, long dataTime) throws FailureException {
		ResultData value = null;
		ReferenceReceiver receiver = null;
		try {
			receiver = new ReferenceReceiver(object, reference.attributeGroup, reference.aspect, receiveOptions, _connection, _userInfo);
			try {
				value = receiver.take(500);
			}
			catch(InterruptedException ignored) {
			}
			receiver.unsubscribe();

			if(value != null) // falls value == null -> finally block
			{
				if(value.getDataTime() > dataTime) {
					value = null;
				}
				else if(reference.getReference() != null && value.hasData()) // nur weitermachen, wenn es auch Daten
				// gibt...
				{
					return takeReference(reference.getReference(), getReference(reference.getAttributeIdent(), value), dataTime);
				}
				else {
					return value;
				}
			}
		}
		catch(FailureException e) {
			e.log();
		}
		finally {
			if(value == null) {
				// Ergebnisdatensatz selbst zusammenbauen, aber ohne Daten
				value = new ResultData(object, new DataDescription(qualifier.attributeGroup, qualifier.aspect), dataTime, null);
			}
		}
		return value;
	}

	/** @param object  */
	@Override
    public void unsubscribe(Object object) {
		collectorThread.abort();
		super.unsubscribe(object);
	}

	/**
	 * Holt eine Refernz aus einem Datensatz.
	 *
	 * @param attributeIdentifier       Attribut (bzw. Attributkette) hinter der der Referenzwert liegt
	 * @param resultData Daten.
	 *
	 * @return Referenziertes Objekt.
	 *
	 * @throws FailureException Fehler beim auslesen der Referenz.
	 */
	private static SystemObject getReference(AttributeIdentifier attributeIdentifier, ResultData resultData) throws FailureException {
		try {
			return attributeIdentifier.getDataElement(resultData.getData()).asReferenceValue().getSystemObject();
		}
		catch(NullPointerException e) {
			throw new FailureException(ErrorMessage.UNEXPECTED_DATA_TYPE + resultData.getData() + ", " + attributeIdentifier, e, LoggerException.WARNING);
		}
	}

	class GetRefThread extends Thread {

		private final OnlineDataReferenceReceiverObject object;

		private final RingBuffer<ResultData> incomingResultData = new RingBuffer<>(100, RingBuffer.UNBOUNDED_SIZE);

		boolean abort = false;

		/** @param object  */
		public GetRefThread(OnlineDataReferenceReceiverObject object) {
			this.object = object;
		}

		/** Beendet den Thread */
		public void abort() {
			abort = true;
			this.interrupt();
		}

		/** @see Runnable#run() */
		@Override
        public void run() {
			ResultData result;
			while(!this.isInterrupted() && !abort) {
				try {
					result = incomingResultData.pop();
					SystemObject so = getReference(qualifier.getAttributeIdent(), result);
					ResultData resultData = takeReference(qualifier.getReference(), so, result.getDataTime());
					object.storeData(resultData);
				}
				catch(FailureException e) {
					e.log();
				}
				catch(InterruptedException e) {
				}
			}
		}

		/**
		 * Weiteren Datensatz hinzufügen
		 *
		 * @param data
		 */
		public void add(ResultData data) {
			incomingResultData.push(data);
		}
	}
}
